# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles prepping and saving the Spock config"""

import os
from uuid import uuid1
from spock._dataclasses import asdict
from spock.handlers import JSONHandler
from spock.handlers import TOMLHandler
from spock.handlers import YAMLHandler
from spock.utils import add_info


class Saver:
    """Base class for saving configs

    Contains methods to build a correct output payload and then writes to file based on the file
    extension

    *Attributes*:

        _writers: maps file extension to the correct i/o handler

    """
    def __init__(self):
        self._writers = {'.yaml': YAMLHandler, '.toml': TOMLHandler, '.json': JSONHandler}

    def __call__(self, *args, **kwargs):
        return Saver()

    def save(self, payload, path, create_save_path=False, extra_info=True, file_extension='.yaml'):  #pylint: disable=too-many-arguments
        """Writes Spock config to file

        Cleans and builds an output payload and then correctly writes it to file based on the
        specified file extension

        *Args*:

            payload: current config payload
            path: path to save
            create_save_path: boolean to create the path if non-existent
            extra_info: boolean to write extra info
            file_extension: what type of file to write

        *Returns*:

            None

        """
        supported_extensions = list(self._writers.keys())
        if file_extension not in list(self._writers.keys()):
            raise ValueError(f'Invalid fileout extension. Expected a fileout from {supported_extensions}')
        # Make the filename
        name = str(uuid1()) + '.spck.cfg' + file_extension
        fid = path / name
        # Fix up values
        out_dict = self._clean_up_values(payload, extra_info, file_extension)
        try:
            if not os.path.exists(path) and create_save_path:
                os.makedirs(path)
            with open(fid, 'w') as file_out:
                self._writers.get(file_extension)().save(out_dict, file_out)
                # yaml.dump(out_dict, file_out, default_flow_style=False)
        except OSError:
            print('Not a valid file path to write to: {}'.format(fid))

    @staticmethod
    def _clean_up_values(payload, extra_info, file_extension):
        """Clean up the config payload so it can be written to file

        *Args*:

            payload: dirty payload
            extra_info: boolean to add extra info
            file_extension: type of file to write

        *Returns*:

            clean_dict: cleaned output payload

        """
        out_dict = {}
        for key, val in vars(payload).items():
            # Append comment tag to the base class and convert the spock class to a dict
            if file_extension == '.json':
                out_dict.update({key: asdict(val)})
            else:
                out_dict.update({('# ' + key): asdict(val)})
        # Convert values
        clean_dict = {}
        for key, val in out_dict.items():
            clean_inner_dict = {}
            for inner_key, inner_val in val.items():
                # Convert tuples to lists so they get written correctly
                if type(inner_val) == tuple:
                    clean_inner_dict.update({inner_key: list(inner_val)})
                elif inner_val is not None:
                    clean_inner_dict.update({inner_key: inner_val})
            clean_dict.update({key: clean_inner_dict})
        if extra_info:
            clean_dict = add_info(clean_dict)
        return clean_dict



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Utility functions for Spock"""

from copy import copy
import os
from time import localtime
from time import strftime
from typing import List
from typing import Tuple
import git
from spock._dataclasses import field


def add_info(out_dict):
    """Adds extra information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    out_dict = add_date_info(out_dict)
    out_dict = add_repo_info(out_dict)
    return out_dict


def add_repo_info(out_dict):
    """Adds GIT information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    try:
        # Assume we are working out of a repo
        repo = git.Repo(os.getcwd())
        out_dict.update({'# Git BRANCH': repo.active_branch.name})
        out_dict.update({'# Git COMMIT SHA': repo.head.object.hexsha})
        if len(repo.untracked_files) > 0 or len(repo.head.commit.diff(None)) > 0:
            git_status = 'DIRTY'
        else:
            git_status = 'CLEAN'
        out_dict.update({'# Git STATUS': git_status})
        out_dict.update({'# Git ORIGIN': repo.remotes.origin.url})
    except git.InvalidGitRepositoryError:
        # But it's okay if we are not
        for key in ('BRANCH', 'COMMIT SHA', 'STATUS', 'ORIGIN'):
            out_dict.update({f'# Git {key}': 'UNKNOWN'})

    return out_dict


def add_date_info(out_dict):
    """Adds date information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    out_dict.update({'# Run Date': strftime('%Y_%m_%d_%H_%M_%S', localtime())})
    return out_dict


def cast(x):
    """Recasts lists as tuples

    *Args*:

        x: object

    *Returns*:

        x: object or object recast as Tuple
    """
    if isinstance(x, list):
        x = tuple(x)
    return x


def _def_list(values: List):
    """Creates a list of default values for List datatype that is mutable

    *Args*:

        values: default list

    Returns:

        list built from default factory

    """
    return field(default_factory=lambda: copy(values))


def _def_tuple(values: Tuple):
    """Creates a tuple of default values for Tuple datatype that is mutable

        *Args*:

            values: default tuple

        Returns:

            tuple built from default factory

        """
    return field(default_factory=lambda: copy(values))



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config decorator that wraps dataclasses"""

from spock._dataclasses import dataclass


def spock_config(*args, **kwargs):
    """Wrapper to dataclass that forms the base of Spock configs

    *Args*:

        *args:
        **kwargs:

    *Returns*:

        frozen dataclass: frozen version of the dataclass
    """
    kwargs['frozen'] = True
    return dataclass(*args, **kwargs)



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

from argparse import Namespace
import argparse
from pathlib import Path
import sys
from typing import _GenericAlias, Generic
from spock._dataclasses import is_dataclass
from spock.payload import Payload
from spock.saver import Saver
from spock.utils import cast


class ConfigArgBuilder:
    """Automatically generates dataclass instances from config file(s)

    This class builds out necessary arguments from *args dataclasses, reads
    the arguments from specified config file(s), and subsequently (via chained
    call to generate) generates each class instance based on the necessary
    field values for each dataclass

    *Attributes*:

        _arg_namespace: generated argument namespace
        _create_save_path: boolean to make the path to save to
        _data_classes: all of the data classes from *args
        _dict_args: dictionary args from the command line
        _optional_types: Dictionary that holds the names of types that are
        optional
        _save_path: list of path(s) to save the configs to

    """
    def __init__(self, *args, **kwargs):
        self._save_path = None
        self._optional_types = {'FloatOptArg', 'IntOptArg', 'StrOptArg',
                                'ListOptArg', 'TupleOptArg', 'SavePathOptArg'}
        self._data_classes = args
        for arg in self._data_classes:
            if not is_dataclass(arg):
                raise TypeError('*arg inputs to ConfigArgBuilder must all be instances of @dataclass')
        self._create_save_path = kwargs.get('create_save_path', False)
        try:
            self._dict_args = self._get_payload(**kwargs)
            self._arg_namespace = self._generate()
        except ValueError as e:
            self.print_usage_and_exit(str(e), sys_exit=False)
            raise ValueError(e)

    def print_usage_and_exit(self, msg=None, sys_exit=True):
        """Prints the help message and exits

        *Args*:

            msg: message to print pre exit

        *Returns*:

            None

        """
        print('USAGE:')
        print(f'  {sys.argv[0]} -c [--config] config1 [config2, config3, ...]')
        print('CONFIG:')
        for data_class in self._data_classes:
            print('  ' + data_class.__name__ + ':')
            dc_vars = vars(data_class)
            for key, val in dc_vars.get('__dataclass_fields__').items():
                if type(val.type).__name__ == 'ChoiceArg':
                    type_name = type(val.type).__name__
                # Magic again -- check for type == type allows for delineation between basic types and list/tuples
                elif type(val.type) == type:
                    type_name = val.type.__name__
                else:
                    type_name = val.type.__origin__.__name__
                    type_name += '[' + self._extract_base_type(val.type).__name__ + ']'
                print(f'    {key}: {type_name}')
        if msg is not None:
            print(msg)
        if sys_exit:
            sys.exit(1)

    def _get_config_paths(self, **kwargs):
        """Get config paths from all methods

        Config paths can enter from either the command line or be added in the class init call
        as a kwarg (configs=[])

        *Args*:

            **kwargs: keyword args

        *Returns*:

            args: namespace of args

        """
        args = self._get_from_arg_parser(**kwargs)
        if kwargs.get('configs') is not None:
            args = self._get_from_kwargs(args, **kwargs)
        return args

    @staticmethod
    def _get_from_arg_parser(**kwargs):
        """Get configs from command line

        Gets the config file(s) from the command line arg parser

        *Args*:

            **kwargs: keyword args

        *Returns*:

            args: namespace of command line args

        """
        # Pull in args via the arg parser pointing to the config file
        parser = argparse.ArgumentParser(description=kwargs.get('desc'), add_help=False)
        parser.add_argument('-c', '--config', required=False, nargs='+', default=[])
        parser.add_argument('-h', '--help', action='store_true')
        args, _ = parser.parse_known_args(sys.argv)
        return args

    @staticmethod
    def _get_from_kwargs(args, **kwargs):
        """Get configs from the configs kwarg


        *Args*:

            args: argument namespace
            **kwargs: keyword args

        *Returns*:

            args: arg namespace

        """
        if type(kwargs.get('configs')).__name__ == 'list':
            args.config.extend(kwargs.get('configs'))
        else:
            raise TypeError('configs kwarg must be of type list')
        return args

    def _get_payload(self, **kwargs):
        """Get the parameter payload from the config file(s)

        Calls the various ways to get configs and then parses to retrieve the parameter payload

        *Args*:

            **kwargs: keyword args

        *Returns*:

            payload: dictionary of parameter values

        """
        args = self._get_config_paths(**kwargs)
        if args.help:
            self.print_usage_and_exit()
        payload = {}
        for configs in args.config:
            payload.update(Payload().payload(self._data_classes, configs))
        return payload

    def __call__(self, *args, **kwargs):
        """Call to self to allow chaining

        *Args*:

            *args:
            **kwargs:

        *Returns*:

            ConfigArgBuilder: self instance
        """
        return ConfigArgBuilder(*args, **kwargs)

    def _generate(self):
        """Method to auto-generate the actual class instances from the generated args

        Based on the generated arguments groups and the args read in from the config file(s)
        this function instantiates the dataclass configs with the necessary field values

        *Returns*:

            auto_dc: namespace containing automatically generated instances of dataclasses
        """
        auto_dc = {}
        for data_classes in self._data_classes:
            dc_build = self._auto_generate(self._dict_args, data_classes)
            auto_dc.update({type(dc_build).__name__: dc_build})
        return Namespace(**auto_dc)

    def generate(self):
        """Generate method that returns the actual argument namespace

        *Returns*:

            argument namespace consisting of all config classes

        """
        return self._arg_namespace

    def save(self, user_specified_path=None, extra_info=True, file_extension='.yaml'):
        """Saves the current config setup to file with a UUID

        *Args*:

            user_specified_path: if user provides a path it will be used as the path to write
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)

        *Returns*:

            self so that functions can be chained
        """
        if user_specified_path is not None:
            save_path = Path(user_specified_path)
        elif self._save_path is not None:
            save_path = Path(self._save_path)
        else:
            raise ValueError('Save did not receive a valid path from: (1) SavePathOptArg or (2) '
                             'the user via user_specified_path')
        # Call the saver class and save function
        Saver().save(self._arg_namespace, save_path, self._create_save_path, extra_info, file_extension)
        return self

    def _auto_generate(self, args, data_class):
        """Builds an instance of a DataClass

        Builds an instance of a dataclass with the necessary field values from the argument
        dictionary read from the config file(s)

        *Args*:

            args: dictionary of arguments read from the config file(s)
            data_class: data class to build

        *Returns*:

            An instance of data_class with correct values assigned to fields
        """
        # Handle the basic data types
        fields = self._handle_basic_arguments(args, data_class)
        return data_class(**fields)

    def _handle_basic_arguments(self, args, data_class):
        """Handles all base types

        These can be easily mapped from the dataclass to another dataclass by some var inspection

        *Args*:

            args: read file arguments
            data_class: instance of a dataclass

        *Returns*:

            fields: dictionary of mapped parameters

        """
        fields = {}
        # Access the vars
        dc_vars = vars(data_class)
        # Get the dataclass name
        dc_name = data_class.__name__
        for key, val in dc_vars.get('__dataclass_fields__').items():
            # pure magic -- Lists, Tuples, etc. are not of type type (they are GenericAlias) so one must
            # check against this before accessing the __name__ attribute which GenericAlias does not have
            if type(val.type) == type:
                if val.type.__name__ == 'SavePathOptArg':
                    self._save_path = args.get(key)
            # Check if namespace is named and then check for key -- checking for local def
            if dc_name in args and key in args[dc_name]:
                fields[key] = self._check_function(args[dc_name][key], val)
            # If not named then just check for keys -- checking for global def
            elif key in args:
                fields[key] = self._check_function(args[key], val)
            # If not found then fallback on defaults if defined
            else:
                default, found_default = self.check_for_defaults(val)
                if not found_default:
                    if type(val.type) == type and val.type.__name__ in self._optional_types:
                        fields[key] = None
                        continue
                    elif type(val.type) != type and val.type.__origin__.__name__ in self._optional_types:
                        fields[key] = None
                        continue
                    elif 'Bool' in val.type.__name__:
                        fields[key] = False
                        continue
                    else:
                        raise ValueError(f'Required value {dc_name}.{key}: no default set or value defined in file')
                fields[key] = self._check_function(default, val)
        return fields

    @staticmethod
    def _int_to_float(inst, target_type):
        """Converts instance int to float

        *Args*:
            inst: instance
            target_type: target type

        *Returns*:
            inst: instance type cast into float

        """
        if target_type == float and type(inst) == int:
            inst = float(inst)
        return inst

    def _check_function(self, x, val):
        """Wrapper around the valid type check with a cast

        *Args*:

            x: instance
            val: value

        *Returns*:

            casted value

        """
        return cast(self._check_valid_type(x, val))

    def _check_valid_type(self, instance, val):
        """Checks that the instance is of the correct type

        *Args*:

            instance: object instance
            val: value

        *Returns*:

            instance: object instance

        """
        if type(val.type) == type:
            # pure magic -- Lists, Tuples, etc. are not of type type (they are GenericAlias) so one must
            # check against this before accessing the __name__ attribute which GenericAlias does not have
            # Get the base variable type
            var_type = val.type.__bases__[0]
            instance = self._int_to_float(instance, var_type)
            valid = isinstance(instance, var_type)
            if not valid:
                raise ValueError(f'Wrong type ({type(instance)}) passed to {val.name}. Require {var_type}')
        elif type(val.type).__name__ == 'ChoiceArg':
            instance = self._check_choice_type(val.type, instance)
            var_type = val.type.set_type
            valid = isinstance(instance, var_type)
            if not valid:
                raise ValueError(f'Wrong type ({type(instance)}) passed to {val.name}. Require {var_type}')
        else:
            # It's an iterable - check it's a list (only iterable provided by markup)
            var_type = self._extract_base_type(val.type)
            iter_name = val.type.__origin__.__name__
            valid = isinstance(instance, (tuple, list))
            if not valid:
                raise ValueError(
                    f'Wrong type ({type(instance).__name__}) passed to {val.name}.\n'
                    f'Require {iter_name}[{var_type.__name__}]')
            instance = tuple((self._int_to_float(i, var_type) for i in instance))
            if len(instance) > 0:
                valid = isinstance(instance[0], var_type)
                if not valid:
                    raise ValueError(
                        f'Wrong type (List[{type(instance[0]).__name__}]) passed to {val.name}.\n'
                        f'Require {iter_name}[{var_type.__name__}]')
        return instance

    @staticmethod
    def _extract_base_type(given_type):
        """Extracts the type from a _GenericAlias

        *Args*:

            tp: type

        *Returns*:

            tp: type of generic type
        """
        if isinstance(given_type, _GenericAlias) and given_type is not Generic:
            return given_type.__args__[0]  # assume we only have generic types with a single argument
        return given_type

    def check_for_defaults(self, val):
        """Checks for default values

        *Args*:

            default: default value
            default_factory: default factory

        *Returns*:

            default_val: value of default
            found_default: boolean if default found

        """
        found_default = False
        default_val = None
        if type(val.type).__name__ == 'ChoiceArg':
            default_val = self._check_choice_type(val.type, val.type.default)
            found_default = True
        elif type(val.default).__name__ != '_MISSING_TYPE':
            default_val = val.default
            found_default = True
        elif type(val.default_factory).__name__ != '_MISSING_TYPE':
            default_val = val.default_factory()
            found_default = True
        return default_val, found_default

    @staticmethod
    def _check_choice_type(choice_set, val):
        """Checks the type and set of a ChoiceArg value

        *Args*:

            choice_set: ChoiceArg instance
            val: value to set
            val_name: name of the parameter

        *Returns*:

            val: value to set

        """
        if val not in choice_set.choice_set:
            raise ValueError(f'{val} is not within the set of defined choices {choice_set.choice_set}')
        return val



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles payloads from markup files"""

from itertools import chain
import os
from pathlib import Path
from spock.handlers import JSONHandler
from spock.handlers import TOMLHandler
from spock.handlers import YAMLHandler


class Payload:
    """Handles building the payload for config file(s)

    This class builds out the payload from config files of multiple types. It handles various
    file types and also composition of config files via a recursive calls

    *Attributes*:

        _loaders: maps of each file extension to the loader class

    """
    def __init__(self):
        self._loaders = {'.yaml': YAMLHandler(), '.toml': TOMLHandler(), '.json': JSONHandler()}

    def __call__(self, *args, **kwargs):
        """Call to allow self chaining

        *Args*:

            *args:
            **kwargs:

        *Returns*:

            Payload: instance of self

        """
        return Payload()

    def payload(self, data_classes, path):
        """Builds the payload from config files

        Main function call that builds out the payload from config files of multiple types. It handles
        various file types and also composition of config files via a recursive calls

        *Args*:

            data_classes: list of dataclasses that define the necessary parameters
            path: path to config file(s)

        *Returns*:

            payload: dictionary of all mapped parameters

        """
        # Match to loader based on file-extension
        config_extension = Path(path).suffix.lower()
        supported_extensions = list(self._loaders.keys())
        if config_extension not in supported_extensions:
            raise TypeError(f'File extension {config_extension} not supported\n'
                            f'Must be from {supported_extensions}')
        # Load from file
        base_payload = self._loaders.get(config_extension).load(path)
        payload = {}
        if 'config' in base_payload:
            payload = self._handle_includes(
                base_payload, config_extension, data_classes, path, payload)
        payload = self._update_payload(base_payload, data_classes, payload)
        return payload

    @staticmethod
    def _update_payload(base_payload, data_classes, payload):
        """Updates the payload

        Checks the parameters defined in the config files against the provided dataclasses and if
        passable adds them to the payload

        *Args*:

            base_payload: current payload
            data_classes: dataclass to roll into
            payload: total payload

        *Returns*:

            payload: updated payload

        """
        # Get basic args
        dc_fields = {dc.__name__: list(vars(dc).get('__dataclass_fields__').keys()) for dc in data_classes}
        # Get the choice args and insert them
        # dc_fields = self._handle_choices(dc_fields, data_classes)
        for keys, values in base_payload.items():
            # check if the keys, value pair is expected by a dataclass
            if keys != 'config':
                # Dict infers that we are overriding a global setting in a specific config
                if isinstance(values, dict):
                    # we're in a namespace
                    # Check for incorrect specific override of global def
                    if keys not in dc_fields:
                        raise TypeError(f'Referring to a class space {keys} that is undefined')
                    for i_keys in values.keys():
                        if i_keys not in dc_fields[keys]:
                            raise ValueError(f'Provided an unknown argument named {keys}.{i_keys}')
                else:
                    # Chain all the values from multiple spock classes into one list
                    if keys not in list(chain(*dc_fields.values())):
                        raise ValueError(f'Provided an unknown argument named {keys}')
            if keys in payload and isinstance(values, dict):
                payload[keys].update(values)
            else:
                payload[keys] = values
        return payload

    def _handle_includes(self, base_payload, config_extension, data_classes, path, payload):  #pylint: disable=too-many-arguments
        """Handles config composition

        For all of the config tags in the config file this function will recursively call the payload function
        with the composition path to get the additional payload(s) from the composed file(s)

        *Args*:

            base_payload: base payload that has a config kwarg
            config_extension: file type
            data_classes: defined dataclasses
            path: path to base file
            payload: payload pulled from composed files

        Returns:

            payload: payload update from composed files

        """
        included_params = {}
        for inc_path in base_payload['config']:
            if not os.path.exists(inc_path):
                # maybe it's relative?
                abs_inc_path = os.path.join(os.path.dirname(path), inc_path)
            else:
                abs_inc_path = inc_path
            if not os.path.exists(abs_inc_path):
                raise RuntimeError(f'Could not find included {config_extension} file {inc_path}!')
            included_params.update(self.payload(data_classes, abs_inc_path))
        payload.update(included_params)
        return payload
